#!/usr/bin/python
#
# Module for doing rpm rollbacks
# Copyright (c) 2000-2002 Red Hat, Inc.
#
# Author: Adrian Likins <alikins@redhat.com>
#


  
#>From python, this "works"
#        rpm.setVerbosity(rpm.RPMLOG_DEBUG)
#        ts = rpm.TransactionSet()
#        print ts.IDTXload()
#        print ts.IDTXglob()
  
#and (untested, this will execute transaction to "transactionid" rollback goal,
#so be careful)
#        ts.Rollback(transactionid)
#
#On to rpmtsSolve() callback bindings ...
 

import rpm
import transaction
import time
import up2dateErrors
from rhpl.translate import _, N_


# ************* NOTE: ************#
# for the sake of clarity, the names "added/removed" as used here
# are indicative of what happened when the original transaction was
# ran. Aka, if you "up2date foobar" and it updates foobar-1-0 with
# foobar-2-0, you added foobar-2-0 and removed foobar-1-0
#
# The reason I mention this explicitly is the trouble of describing
# what happens when you rollback the transaction, which is basically
# the opposite, and leads to plenty of confusion
#


class RollbackTransaction:
    def __init__(self):
        # these list are lists of tuples of (hdr, filename or instance)
        self.addedPackages = []
        self.removedPackages = []
        self.tid = None

    def getData(self):
        data = {}
        removedList, addedList = self.genNames()
        data['added'] = addedList
        data['removed'] = removedList
        data['tid'] = self.tid
        return data

    # method to generate a datastructure useful for transporting
    #over the wire, and/or displaying to the user
    def genNames(self):
        data = {}
        removedList = []
        addedList = []
        for (hdr,instance) in self.removedPackages:
            epoch = ""
            if hdr['epoch']:
                epoch = "%s" % hdr['epoch']
            pkgInfo = (hdr['name'],hdr['version'],hdr['release'],epoch, hdr['arch'], "-")
            removedList.append(pkgInfo)
        for (hdr,filename) in self.addedPackages:
            epoch = ""
            if hdr['epoch']:
                epoch = "%s" % hdr['epoch']
            pkgInfo = (hdr['name'],hdr['version'],hdr['release'],epoch, hdr['arch'],  "+")
            addedList.append(pkgInfo)
        return (removedList, addedList)

    def display(self):
        out = ""
        out = out +  "install time: %s\t tid:%s\n" % (time.ctime(self.tid), self.tid)
        removedList, addedList = self.genNames()
        for pkgInfo in removedList:
            out = out + "\t\t[%s] %s-%s-%s:%s\n" % (pkgInfo[5], pkgInfo[0], pkgInfo[1], pkgInfo[2], pkgInfo[3])
        for pkgInfo in addedList:
            out = out + "\t\t[%s] %s-%s-%s:%s\n" % (pkgInfo[5], pkgInfo[0], pkgInfo[1], pkgInfo[2], pkgInfo[3])
        return out

    # generate a rpm transaction that we can turn around and run
    def genTransaction(self):
        self.ts = rpm.TransactionSet()
        self.ts.setVSFlags(~(rpm.RPMVSF_NOMD5))

        #FIXME: Insert policy issues here
        #  possibilities:
        #     max size of transactions
        #     pkg/file list to never rollback
        #     date to never rollback past
        #     protected tids?

        # remove the added packages and
        # install the removed packages. That almost
        # makes sense
        for (hdr,instance) in self.addedPackages:
#            print "name: %s instance: %s" % (hdr['name'], instance)
            self.ts.addErase(instance)
        for (hdr, filename) in self.removedPackages:
#            print "name: %s filename: %s" % (hdr['name'], filename)
            self.ts.addInstall(hdr,hdr, 'i')

        deps = self.ts.check()
        if deps:
            raise up2dateErrors.DependencyError(_("Dependencies should have already been resolved, but they are not."), deps)

        # turn off md5/rsa/dsa
        self.ts.setVSFlags(rpm.RPMVSF_NODSA|rpm.RPMVSF_NOMD5|rpm.RPMVSF_NORSA|rpm.RPMVSF_NEEDPAYLOAD)

        # FIXME: self.ts.addTSFlags (coming to bindings) to persistently set
        # stuff like "--oldpackage"
        self.ts.setFlags(rpm.RPMPROB_FILTER_OLDPACKAGE|rpm.RPMTRANS_FLAG_NOMD5)
        return self.ts

class Rollback:
    def __init__(self, ts = None):
        if ts:
            self.ts = ts
        else:
            self.ts = transaction.initReadOnlyTransaction()
        self.ts.pushVSFlags(-1)
        self.__genAvailableTidHdrDict()
        self.__genInstalledTidHdrDict()
        self.__genTransDict()

    def __genTransDict(self):
        # generate a dict with tid as the key and a "transaction" object
        # as the value
        self.transDict = {}
        for tid in self.availableTidHdrDict.keys():
            # need has_key?
            rollbackTransaction = RollbackTransaction()
            if self.installedTidHdrDict.has_key(tid):
                rollbackTransaction.addedPackages = self.installedTidHdrDict[tid]
            if self.availableTidHdrDict.has_key(tid):
                rollbackTransaction.removedPackages = self.availableTidHdrDict[tid]
            rollbackTransaction.tid = tid
            self.transDict[tid] = rollbackTransaction

    # this is a map of tid -> list of packages installed with that tid
    def __genInstalledTidHdrDict(self):
        self.installedTidHdrDict = {}
        tidlist = self.ts.IDTXload()
        for (tid, hdr, instance) in tidlist:
            if self.installedTidHdrDict.has_key(tid):
                self.installedTidHdrDict[tid].append((hdr,instance))
            else:
                self.installedTidHdrDict[tid] = [(hdr, instance)]

    def __genAvailableTidHdrDict(self):
        # this generates of tids that we can actually rollback to,
        # ie, the ones on disk
        self.tidHdrDict = {}
        self.availableTidHdrDict = {}
        transList = self.ts.IDTXglob()
        for (tid, hdr, filename) in transList:
            if self.availableTidHdrDict.has_key(tid):
                self.availableTidHdrDict[tid].append((hdr,filename))
            else:
                self.availableTidHdrDict[tid] = [(hdr,filename)]

    def __getitem__(self, item):
        return self.availableTidHdrDict[item]


    def getPackages(self, item):
        hdrlist = self.availableTidHdrDict[item]
        tmplist = []

        return tmplist
            
    def keys(self,item):
        return self.availableTidHdrDict.keys()

    def values(self,item):
        return self.availableTidHdrDict.values()

    def has_key(self, key):
        return self.availableTidHdrDict.has_key(key)
    
    def printTransactions(self):
        tidlist = self.availableTidHdrDict.keys()
        tidlist.sort()
        for tid in tidlist:
            print self.showTransaction(tid)
            
    def showTransaction(self, tid):
        if self.transDict.has_key(tid):
            return self.transDict[tid].display()

    # return a list of transactions as data structs
    def getTransactionsData(self):
        data = {}
        for tid in self.transDict.keys():
            data["%s" % tid] = self.transDict[tid].getData()
        return data
    
    def getLatestTid(self):
        tidlist = self.availableTidHdrDict.keys()
        tidlist.sort()
        return tidlist[-1]    

    
    def doRollback(self,tid):
        self.ts.rollback(tid)

    # this *Info methods are for actions
    def getTidInfo(self,tid):
        packages = self.getPackages(tid)
        infodict = {'tid': tid, 'packages':packages}
        return infodict

    def getUndoInfo(self):
        tid = self.getLatestTid()
        return self.getTidInfo(tid)
    
    def previewUndo(self):
        tid = self.getLatestTid()
        return self.showTransaction(tid)
    
    def undo(self):
        tid = self.getLatestTid()
        self.doRollback(tid)


def main():
    import sys

#    rpm.setVerbosity(rpm.RPMLOG_DEBUG)
    # let me rollback deletions
    rpm.addMacro("_unsafe_rollbacks", "1")
    rollback = Rollback()
    foo = rollback.getLatestTid()

    print rollback.getTransactionsData()
    print foo
    print
    print rollback.printTransactions()
    print
    sys.exit()
    print "rolling back tid: %s with packages: %s" %  (foo, rollback.getPackages(foo))
    print
    bleah = rollback.transDict[foo]
    print bleah.display()

    import pprint
    pprint.pprint(bleah.getData())

#    ts = bleah.genTransaction()
#    print rollback.undo()
#    import up2date
#    import wrapperUtils
#    print ts
#    rpmCallback = wrapperUtils.RpmCallback()
#    print up2date.runTransaction(ts,rpmCallback.callback)


if __name__ == "__main__":
    main()


