#!/usr/bin/python -u
#
# Client code for Update Agent
# Copyright (c) 1999-2002 Red Hat, Inc.  Distributed under GPL.
#
# Author: Preston Brown <pbrown@redhat.com>
#         Adrian Likins <alikins@redhat.com
#         Cristian Gafton <gafton@redhat.com>
#
"""
Guts of the up2date client. This module provides abstration
of all the remote rpc calls that the client needs to make, includes
the dependency solve, and other assorted bits.
"""

import os,sys
import string
import re
import fnmatch, glob
import rpm
import socket
#import rhnxmlrpclib as xmlrpclib
import time
import StringIO
import up2dateUtils
import up2dateLog
import up2dateErrors
import up2dateAuth
import rpmSource
import rhnDefines
import rhnPackageInfo
import rhnChannel
import checkbootloader
import xmlrpclib
import rpmUtils
import rpcServer
import depSolver
import headers
import rollbacks
import transaction

# so we can use the same code with python 1.5.2 and 2.2
try:
    from rhn import rpclib
except ImportError:
    rpclib = __import__("xmlrpclib")
    

from rhpl.translate import _, N_

import config

cfg = config.initUp2dateConfig()

_ts = transaction.initReadOnlyTransaction()

def reloadConfig():
    cfg.load()

def version():
    # substituted to the real version by the Makefile at installation time.
    return "3.0.7.2"

# get around cached dependencies.  Gross ugly hack.
#rpm.addMacro("_dbi_tags","Packages:Name:Basenames:Group:Requirename:Providename:Conflictname:Triggername")

# turn off epoch promotion for the time being
rpm.setEpochPromote(0)

# this is a "gloabal" variable caching the list of all available packages
global avail_package_list
avail_package_list = []

try:
    log = up2dateLog.initLog()
except IOError, e:
    print "Unable to open log file. The error was: %s" % e
    print _("You must run the Update Agent as root.")
    sys.exit(1)

def checkForStorageDir():
    dir = cfg["storageDir"]
    if os.access(dir, os.X_OK):
        return 0
    else:
        return 1

headerCache = headers.HeaderCache()

def getAvailablePackageList(msgCallback=None, progressCallback=None):
    global avail_package_list


    if len(avail_package_list):
        return avail_package_list

    package_list = rhnPackageInfo.availablePackageList(
        msgCallback, progressCallback)

    availPkgsDict = {}
    #tmp_args,tmp_method = xmlrpclib.loads(data)
    for pkg in package_list:
        archScore = rpm.archscore(pkg[4])
        if archScore <= 0:
            # incompatible architecture
            continue

        pkgName = pkg[0]
        if not availPkgsDict.has_key(pkgName):
            availPkgsDict[pkgName] = pkg
            continue

        # Already seen this name
        ret = up2dateUtils.comparePackages(availPkgsDict[pkgName], pkg)
        if ret > 0:
            # don't care, we already have a better version
            continue
        if ret < 0:
            # Better version
            availPkgsDict[pkgName] = pkg
            continue
        # Same version, so compare the arches
        oldScore = rpm.archscore(availPkgsDict[pkgName][4])
        if oldScore <= archScore:
            continue
        # Package with a better architecture
        availPkgsDict[pkgName] = pkg

    # Sort the keys (i.e. the package names)
    avail_package_list = availPkgsDict.values()
    avail_package_list.sort()
    return avail_package_list

def getPackageSearchPath():
    dir_list = []
    dir_list.append(cfg["storageDir"])

    dir_string = cfg["packageDir"]
    if dir_string: 
    	dir_list = dir_list + string.split(dir_string, ':')
    return dir_list




# needed for actions, maybe client to at some point
def getErrataInfo(errata_id):
    s = rpcServer.getServer()
    packages = rpcServer.doCall(
        s.errata.getErrataInfo, up2dateAuth.getSystemId(), errata_id)
            
    return packages


# pass the lite flag if you only want the "lite" header
def getHeader(pkg,lite=None):
    hdr = None

    hdr,type = rpcServer.doCall(rpmSource.psc.headerSource.getHeader, pkg)
    return hdr
    
    

# implicitly places headers in the headerCache.
# FIXME: i think this can die now with headers.HeaderList
def getHeaders(pkgList, msgCallback = None, progressCallback = None,
               refreshCallback = None):

    headerList = []
        
    # if we get a single header, listify it
    # double check to make sure it's a list of lists and not just
    # a single list item
    if type(pkgList) == type([]):
        if len(pkgList):
            if type(pkgList[0]) == type(""):
                tmp = [pkgList]
                pkgList = tmp
        else:
            return []
    else:
        raise TypeError, "pkgList must be a list of lists"


    # munge the list if we have the size include.
    # xml-rpc doesnt seem fond of that being included
    total = len(pkgList)

    # reset progress bar
    if progressCallback:
        progressCallback(0, total)

    if not refreshCallback:
        refreshCallback = 0

    for pkg in pkgList:
        if refreshCallback:
            refreshCallback()
        hdr = None
        hdr, sourcetype = rpcServer.doCall(
            rpmSource.psc.headerSource.getHeader, pkg)
        
        headerList.append(hdr)

        if progressCallback:
            progressCallback(len(headerList), total)
 
    return headerList

def removeHeader(pkg):
    # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.hdr" %
                          (cfg["storageDir"],
                           up2dateUtils.pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)

    fileNames = glob.glob("%s/%s.noarch.hdr" %
                          (cfg["storageDir"],
                           up2dateUtils.pkgToString(pkg)))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)


def isPackageCached(pkg):

    baseFileName = "%s-%s-%s.%s.rpm" % (pkg[0], pkg[1], pkg[2], pkg[4])

    fileName = "%s/%s" % (cfg["storageDir"],
                          baseFileName)

    if (os.access(fileName, os.R_OK)):
        size = os.stat(fileName)[6]
        if size == string.atol(pkg[5]):
            return 1
    return 0


def getPackage(pkg, msgCallback = None, progressCallback = None):

    buffer = rpcServer.doCall(rpmSource.psc.packageSource.getPackage, pkg, msgCallback, progressCallback)

    # retrieve source packages as well
    if cfg['retrieveSource']:
        hdr = getHeader(pkg)
        srcrpm = hdr['sourcerpm']
        channel = pkg[6]
        buffer = rpcServer.doCall(rpmSource.psc.sourcePackageSource.getPackageSource,
                        channel,
                        srcrpm,
                        msgCallback,
                        progressCallback)
        
    return 0


def removePackage(pkg):
     # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.*" %
                          (cfg["storageDir"],
                           up2dateUtils.pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
            
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)
        
    fileNames = glob.glob("%s/%s.noarch.*" %
                          (cfg["storageDir"],
                           up2dateUtils.pkgToString(pkg)))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
        except:
            raise up2dateErrors.FileError(_("cannot remove %s") % fileName)


def avail_package_dict():
    # returns a hash version of the Avialable Package List
    availablePkgList = getAvailablePackageList()
    pkgDict = {}
    for pkg in availablePkgList:
        pkgDict[up2dateUtils.pkgToStringArch(pkg)] = pkg
    return pkgDict


def getArchesOfAvailablePackages(availablePkgList,pkgName):
    archList = []
    for pkg in availablePkgList:
        if pkgName == pkg[0]:
            archList.append(pkg[4])

    return archList

# convience class to make removeInstalledPackagesFromList cleaner
class UpdateList:
    def __init__(self):
        self.updateDict = {}
        self.__updateIndex = None

    def nameList(self):
        return self.updateDict.keys()

    # gen a list of just the name,version,release,epoch, and arch fields
    def nvreaList(self):
        return map(lambda a: a[:5], self.updatedList())

    def nvreList(self):
        return map(lambda a: a[:4], self.updatedList())

    def appendIfNewer(self, installedPkg, pkg):
        pkgName = pkg[0]
        # do we already have a package like this?
        if not self.updateDict.has_key(pkgName):
            self.updateDict[pkgName] = pkg
            return
        # do we have a match for name-version-release
        if pkg[:4] in self.nvreList():
            # see if this is a better matching architecture
	    archScore = rpm.archscore(pkg[4])
            # incompatible arch
            if archScore <= 0:
                return
	    oldScore = rpm.archscore(self.updateDict[pkgName][4])
            if archScore < oldScore: # retain the better arch
                self.updateDict[pkgName][4] = pkg[4]
            return
        # else, we might gave an older one that needs updating
        ret = up2dateUtils.comparePackages(self.updateDict[pkgName],pkg)
        if ret > 0: # if pkg is newer
            self.updateDict[pkgName] = pkg

    def updatedList(self):
        return self.updateDict.values()

    def delete(self, pkg):
        del self.updateDict[pkg[0]]



# check if the signature matches the redhat.com one
def hasBadSignature(pkg):
    hdr = getHeader(pkg)
    fileName = "%s/%s.%s.rpm" % (cfg["storageDir"],
                                 up2dateUtils.pkgToString(pkg),
                                 hdr['arch'])

    fd = os.open(fileName, os.O_RDONLY)
    _ts.setVSFlags(0)
    try:
        hdr = _ts.hdrFromFdno(fd)
    except rpm.error, e:
        if str(e) == "public key not availaiable":
            return 1
        if str(e) == "public key not available":
            return 1
        if str(e) == "public key not trusted":
            return 3
        if str(e) == "error reading package header":
            return 2

    if hdr['SIGGPG'] == None:
	return 1
    os.close(fd)    
    # all the gory bits of checking the sig is in rpmUtils now
    return 0


# The new dryRun
def dryRun(pkgList, msgCallback = None, progressCallback = None,
           refreshCallback=None):
    log.log_debug("Called dryRun", pkgList)
        
    availPkgs = getAvailablePackageList()
    depsolve = depSolver.DependencySolver(pkgList, availPkgs)

    # set the callbacks wwe need
    depsolve.set_callback("msg", msgCallback)
    depsolve.set_callback("progress", progressCallback)
    depsolve.set_callback("refresh", refreshCallback)
    
    # set up the trabsaction set & all
    depsolve.setup()
    while 1:
        log.log_debug("Selected so far", depsolve.selected)
        ret = depsolve.solvedep()
        log.log_debug("Got code %d from solvedep" % ret)
        if ret == 0:
            break
        elif ret < 0: # not really reached
            raise up2dateErrors.DependencyError(_(
                "Could not complete dependency resolution"))
    rc = depsolve.run()
    if progressCallback:
        progressCallback(1, 1)       
    if len(rc):
        msg = ""
        for code in rc:
            msg = msg + code[0] + "\n"
            raise up2dateErrors.ConflictError(
                _("Test install failed because of package conflicts:\n") + msg,
                rc)
    newList = []
    for s in depsolve.selected:
        if not s in pkgList:
            pkgList.append(s)
            newList.append(s)
    return (newList, [])

def undo():
    # FIXME: instrument this with better error handling
    rollback = rollbacks.Rollback()
    print "Rolling back the packages installed:" 
    print rollback.previewUndo()
    ret = 0
    rollback.undo()
#    except:
#        print "rollback failed"
#        ret = 1
    return ret

def genTransaction(transactionData):
    # generate a rpm.TransactionSet from a TransactionData object
    #

    ts = rpm.TransactionSet()

    pkgList = transactionData.data['packages']

    kernels = []
    for (pkg,mode) in pkgList:
        if mode == 'u' or mode == 'i':
            hdr = getHeader(pkg)
            if not hdr:
                raise up2dateErrors.RpmError(_(
                    "Could not find header for %s") % pkg)
            if "kernel" in hdr['Providename']:
                kernels.append(hdr)
            ts.addInstall(hdr,hdr, mode)
        elif mode == "e":
            instances = rpmUtils.installedHeaderIndexByPkg(pkg)
            for instance in instances:
                ts.addErase(instance)
        else:
            # dunno what this is, but we might use it
            ts.addInstall(hdr, hdr, mode)

    if transactionData.data.has_key('vsflags'):
        ts.setVSFlags(transactionData['vsflags'])

    return ts

    
def runTransaction(ts, rpmCallback):
    oldInstalledPackages = rpmUtils.getInstalledPackageList()
     # JBJ: probabaly do not need this ordering
#    ts.order()
    deps = ts.check()
    if deps:
        raise up2dateErrors.DependencyError(_(
            "Dependencies should have already been resolved, "\
            "but they are not."), deps)

    if cfg["justdb"]:
        ts.setFlags(rpm.RPMTRANS_FLAG_JUSTDB)
#        ts.setProbFilter(0)
        rc = ts.run(rpmCallback, cfg["storageDir"])
    else:
        # ( 1 << 10) == rpm.RPMTRANS_FLAG_REPACKAGE
        transflag = 0
	if cfg["enableRollbacks"]:
	    log.log_me("Creating rollback packages...")
	    transflag = rpm.RPMTRANS_FLAG_REPACKAGE 

        ts.setFlags(transflag)
#        ts.setProbFilter(0)
        rc = ts.run(rpmCallback, cfg["storageDir"])
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise up2dateErrors.RpmError(_(
            "Failed running transaction of  packages: %s") % errors)

    newInstalledPackages = rpmUtils.getInstalledPackageList()

    removedPackages = []
    addedPackages = []

    # find packages that were removed
    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    for pkg in newInstalledPackages:
        if pkg not in oldInstalledPackages:
            addedPackages.append(pkg)

    if len(removedPackages):
    	rhnPackageInfo.remoteDelPackages(removedPackages)
    if len(addedPackages):    
    	rhnPackageInfo.remoteAddPackages(addedPackages)

# FIXME: untested
def installPackages(pkgList, rpmCallback):
    try:
        log.log_me("installing packages: %s " % up2dateUtils.pprint_pkglist(pkgList))
    except:
        pass

    # a list of all the packge names to treat as kernels.
    # "hardcode" it here so people dont mess with it as much
    total = len(pkgList)

    # delete any existing db, open in write mode
    ts = rpm.TransactionSet()
    ts.setVSFlags(8)


    kernels = []
    for p_i in range(total):
        hdr = getHeader(pkgList[p_i])
        if hdr:
            if "kernel" in hdr['Providename']:
                ts.addInstall(hdr,hdr, "i")
                kernels.append(hdr)
            else:
                ts.addInstall(hdr, hdr, "u")
        else:
            raise up2dateErrors.RpmError(_(
                "Could not find header for %s") % pkgList[p_i])

    runTransaction(ts, rpmCallback)

    return kernels

def removePackages(pkgList, rpmCallback):
    try:
        log.log_me("removing packages: %s " % up2dateUtils.pprint_pkglist(pkgList))
    except:
        pass  

    total = len(pkgList)

    

    # delete any existing db, open in write mode
    ts = rpm.TransactionSet()
    ts.setVSFlags(8)

    # FIXME: just that a package is installed before claiming
    # it cant be removed

    remove_list = []
    for p_i in range(total):
        for pattern in cfg["removeSkipList"]:
            pkgName = "%s-%s-%s" % (pkgList[p_i][0],pkgList[p_i][1],pkgList[p_i][2])
            # check the package name against the globs in removeSkipList
            if fnmatch.fnmatch(pkgList[p_i][0], pattern):
                raise up2dateErrors.RpmRemoveSkipListError, pkgName
            else:
                remove_list.append(pkgName)

    failed_packages = {}
    for pkgName in remove_list:
        try:
            ts.addErase(pkgName)
        except rpm.error, e:
            failed_packages[pkgName] = e

    # if we get failed packages, raise the hash
    # of failed packages
    if len(failed_packages.keys()):
        raise up2dateErrors.RpmRemoveError, failed_packages
    
    runTransaction(ts,rpmCallback)
    
    return 0

def installBootLoader(kernels):
    
    # option to disable this permanently for those using grub,etc
    if cfg["noBootLoader"]:
        return

    # dont run the new lilo if were running "justdb"
    if cfg["justdb"]:
        return

    log.log_me("Modifying bootloader config to include the new kernel info")
    kernelList = []
    # annoying, but kernels provide "kernel", "kernel-drm", "kernel-dfasdf", etc 
    for krn in kernels:
        if "kernel-BOOT" in krn['Providename']:
            extraInfo = "kernel-BOOT"
        elif "kernel-smp" in krn['Providename']:
            extraInfo = "kernel-smp"
        elif "kernel-enterprise" in krn['Providename']:
            extraInfo = "kernel-enterprise"
        else:
            extraInfo = "kernel"

        # this logics a bit weird
        if extraInfo == None:
            infoString = ""
        elif extraInfo == "kernel":
            infoString = ""
        elif extraInfo == "kernel-BOOT":
            infoString = "BOOT"
        elif extraInfo == "kernel-enterprise":
            infoString = "enterprise"
        elif extraInfo == "kernel-smp":
            infoString = "smp"
        else:
            infoString = ""


        verRel = "%s-%s%s" % (krn['version'], krn['release'],infoString)
        kernelList.append((verRel,extraInfo))
        log.log_me("Adding %s to bootloader config" % verRel)


    bootloader = checkbootloader.whichBootLoader()
    
    if up2dateUtils.archGlob() == 'i?86' and bootloader == "LILO":
        __install_lilo(kernelList)
    elif up2dateUtils.archGlob() == 'i?86' and bootloader == "GRUB":
        # at the moment, kernel rpms are supposed to grok grub
        # and do the Right Thing. Just a placeholder for doc purposes
        __install_grub(kernelList)
    elif up2dateUtils.archGlob() == "ia64":
        __install_elilo(kernelList)

def __install_elilo(kernelList):
    import elilocfg
        
    log.log_me("Running elilo with the new configuration")
    ret = elilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_lilo(kernelList):
    import lilocfg
    
    log.log_me("Running lilo with the new configuration")
    ret = lilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_grub(kernelList):
    import grubcfg

    log.log_me("Installing the kernel via grub")
    ret  = grubcfg.installNewImages(kernelList,test=0)
    return ret



def test():

    # login and whatnot
    up2dateAuth.updateLoginInfo()
    
    # lets build a transaction that removes all the "kde-*" packages
    # and adds all the gnome* packages
    removeHdrs = rpmUtils.installedHeaderByKeyword(name="kde*")

    
    # create a TransactionData object
    tsd = transaction.TransactionData()

    for hdr in removeHdrs:
        # stringify epoch
        if hdr['epoch']:
            epoch = str(hdr['epoch'])
        else:
            epoch = "0"
        pkg = [hdr['name'], hdr['version'],
               hdr['release'], epoch,
               hdr['arch']]
        mode = 'e'
#        print "blip: %s" % pkg
        tsd.data['packages'].append((pkg, mode))

    import fnmatch
    pkgs = getAvailablePackageList()
    for pkg in pkgs:
        if fnmatch.fnmatch(pkg[0], "gnome*"):
            mode = 'i'
            print "blargh: (%s,%s)" % (pkg,mode)
            tsd.data['packages'].append((pkg,mode))

    print tsd
    print tsd.display()

    tsi = genTransaction(tsd)
    print tsi

    # for callback
    import wrapperUtils
    rpmCallback = wrapperUtils.RpmCallback()
    runTransaction(tsi, rpmCallback)
    

if __name__ == "__main__":
    test()
